#include <iostream>


/*
  Un petit programme pour illustrer la maniere dont sont appelles implicitement
  les constructeurs, le destructeur et l'operateur "=" d'une classe.

  Ici, on definit une classe sans interet (elle n'a pas de membres), mais dans
  laquelle les constructeurs, le destructeur et l'operateur= se "manifestent"
  quand il sont appeles. Cela permet de voir a quel moment de l'execution d'un
  programme ces fonctions sont appelees.
*/


class ma_classe
{
public:
  ma_classe(void)
  /*
	Fonction de meme nom que la classe, sans type de retour. C'est un constructeur.
	Par ailleurs, n'ayant pas d'argument, il s'agit du constructeur par defaut de la
	classe "ma_classe", qui sera appele a chaque declaration du type "ma_classe x;".
  */
  {
	std::cout << "Constructeur par defaut\n";
  }

  ma_classe(ma_classe &)
  /*
	Un autre constructeur, prenant comme argument un reference sur
	un objet de la classe. Il s'agit donc du constructeur *par copie*.
	En temps normal, ce constructeur sert a recopier le contenu
	d'une variable dans une autre variable.
  */
  {
	std::cout << "Constructeur par copie\n";
  }

  ma_classe& operator= (ma_classe&)
  // L'operateur =.
  {
	std::cout << "Operateur =\n";
  }
  
  ~ma_classe(void)
  // Le destructeur, appele a chaque fois qu'une variable disparait.
  {
	std::cout << "Destructeur\n";
  }
};



void f(ma_classe x)
{
  // Dans cette fonction, on prend comme argument une variable de type ma_classe.
  // Par consequent, cette fonction va recopier cet argument dans une variable locale,
  // qui sera detruite a la fin de la fonction. Il y aura donc un appel du constructeur
  // par copie et un appel du destructeur.
  std::cout << "Dans f\n";
}

void g(ma_classe &x)
{
  // Ici, on prend comme argument une *reference* sur un objet de type ma_classe.
  // Il n'y a donc pas de recopie, pas d'appel de constructeur, pas d'appel du destructeur.
  std::cout << "Dans g\n";
}

void h(void)
{
  // Cette fonction contient une variable locale de type ma_classe. Cette variable
  // va etre creee par un appel du constructeur, puis sera detruite a la fin de
  // l'execution de la fonction.
  std::cout << "Debut de h\n";
  ma_classe x;
  std::cout << "Fin de h\n";
}



int main(void)
{

  std::cout << "---- Declaration de deux variables ----\n";
  ma_classe x, y; // On utilise ici le constructeur par defaut.
  std::cout << "\n\n";

  std::cout << "---- Affectation de y a x ----\n";
  x = y; // On utilise l'operateur =.
  std::cout << "\n\n";

  std::cout << "---- Declaration de z ----\n";
  ma_classe z(x); // Cette syntaxe est un appel de constructeur. Il s'agit ici du constructeur par copie.
  std::cout << "\n\n";

  std::cout << "---- Declaration de w ----\n";
  ma_classe w=x;
  // Malgre les apparences, il s'agit ici d'un appel du constructeur par copie.
  // Cette syntaxe n'est pas equivalente a "ma_classe w; w=x;". Dans ce dernier cas,
  // on a un appel du constructeur par defaut et un appel de l'operateur =.
  std::cout << "\n\n";

  std::cout << "---- Appel de f ----\n";
  f(x);
  std::cout << "\n\n";

  std::cout << "---- Appel de g ----\n";
  g(x); 
  std::cout << "\n\n";

  std::cout << "---- Appel de h ----\n";
  h(); 
  std::cout << "\n\n";

  std::cout << "---- Fin ----\n";

  // Fin de la fonction, les variables creees vont etre
  // detruite les unes apres les autres.
  
  return 0;
}
